#include "gamelib.h"

#include <stdarg.h>

void	init_gamelib()
{
#ifdef	INIT_MERSENA
	time_t gt;
        time(&gt);

        init_genrand((unsigned long)gt);
#endif
}


int 
	playerCounter = 0, 
	entityCounter = 0;


int	newPlayerId() { return playerCounter++; }
int	newEntityId() { return entityCounter++; }

real    myrand() {
        return genrand_real1();
}


real    uniform(real f, real t)
{
        real k = myrand();
        return (t-f)*k + f;
}


int     int_uniform(real f, real t)
{
        return (int) uniform(f, t);
}


real	procenta(real base, real pr)
{
	assert(pr>=0 && pr<=100);
	return base/100.0 * pr;
}


real	vahovane(real m1, real c1, real m2, real c2)
{
	if (m1+m2==0) return 0;
	return (m1*c1+m2*c2)/(m1+m2);
}


void	gprintf(int lev, const char *fmt, ...)
{
	return ;
	va_list list;
	va_start(list, fmt);
	vprintf(fmt, list);
	va_end(list);
}


void	Strategy::print()
{
	for (int a=0; a<size; a++) 
		zprintf("%.0f ", vals[a]);
	zprintf("\n");
}

void	Strategy::output(FILE *sb)
{
	fprintf(sb, "%d ", order);
	for (int a=0; a<size; a++) 
		fprintf(sb, "%.0f ", vals[a]);
	fprintf(sb, "\n");
}


bool	Strategy::input(FILE *sb)
{

	fscanf(sb, "%d ", &order);
	if (feof(sb)) return false;
	float v;

	for (int a=0; a<size; a++) {
		fscanf(sb, "%f ", &v);
		vals[a]=v;
		//printf("Vala %d %.0f %d\n", order, v, a);
	}

	return true;
}


void	Profile::become_num(longcis eenum)
{
	num = eenum;
	gamed->map2_prof(eenum, this);

	load_strats();
}


void	Profile::become_from_addr()
{
	num = gamed->map2_num(this);

	load_strats();
}


void	Profile::load_strats()
{
	for (int a=0; a<gamed->N; a++) 
		strats[a] = gamed->get_strat(a, Addr[a]);
}


void	Profile::printAddr(bool nl, FILE *sb) {
	//zprintf("[%d]:", gamed->N);
	//std::cout << num << " ";

	for (int a=0; a<gamed->N; a++) fprintf(sb, "%d ", Addr[a]);
	if (nl) fprintf(sb,"\n");
}

bool	Profile::readAddr(FILE *sb)
{
	for (int a=0; a<gamed->N; a++) {
		fscanf(sb, "%d", &Addr[a]);
		if (feof(sb)) return false;
	}

	become_from_addr();
	return true;
}


void	Profile::printU(bool nl, FILE *sb)
{
	for (int a=0; a<gamed->N; a++) fprintf(sb, "%.0f ", U[a]);
	if (nl) fprintf(sb, "\n");
}



void	Profile::reset_u() {
	for (int a=0; a<gamed->N; a++) U[a]=0;
}


bool	Profile::operator == (Profile &prf)
{
	if (prf.gamed!=gamed) return false;
	for (int a=0; a<gamed->N; a++) if (Addr[a]!=prf.Addr[a]) return false;
	return true;
}


void	Profile::copy(Profile &prf) {
	gamed = prf.gamed;
	num = prf.num;
	for (int a=0; a<gamed->N; a++) {
		Addr[a]=prf.Addr[a];
		U[a]=prf.U[a];
		strats[a]=prf.strats[a];
	}
}


void	Profile::copy_u(Profile &prf)
{
	for (int a=0; a<gamed->N; a++) {
		U[a]=prf.U[a];
	}
}


void	GameDescription::map_init()
{
	int n,a;

	for (n=0; n<N; n++) 
		strats_size[n]=strats[n].size();

	longcis i=1;
	//std::cout << "map_init: " << i << "\n";

	for (a=N-1; a>=0; a--) {
		idxs[a]=i;
		//std::cout << "map_init: " << i << "\n";
		i *= strats_size[a];
		//zprintf("%Lg ", idxs[a]);
	}
}


void	GameDescription::map2_prof(longcis num, Profile *kam)
{
	int a;
	longcis pom;

	for (a=0; a<N-1; a++) {
		pom = num / idxs[a];
		kam->Addr[a] = pom.get_si();
		num -= kam->Addr[a] * idxs[a];
	}

	kam->Addr[a]=(int) num.get_si();
}


longcis	GameDescription::map2_num(Profile *kam)
{
	longcis i=0;
	for (int a=0; a<N; a++) 
		i += ((longcis) kam->Addr[a]) * idxs[a];
	return i;
}



void	GameDescription::save_game_format_iter(Profile *prof, int level, FILE *sb, Ucache_manager &ucan)
{
	if (level == prof->gamed->N) {
		prof->become_from_addr();
		Profile *on = ucan.getp(prof->num);
		real	co;

		for (int pl=0; pl<prof->gamed->N; pl++) {
			co = (on) ? on->U[pl] : 0;
			fprintf(sb, "%.0f ", co);
		}

		return ;
	}

	int	maxa = prof->gamed->nstrats(level);

	for (prof->Addr[level]=0; prof->Addr[level]<maxa; prof->Addr[level]++) {
		save_game_format_iter(prof, level+1, sb, ucan);
	}
}


void	GameDescription::save_game_nfg(const char *fn, Ucache_manager &ucan)
{
	FILE	*sb = fopen(fn, "wt");
	assert(sb!=0);

	fprintf(sb, "%d\n", N);
	for (int pl=0; pl<N; pl++) fprintf(sb, "%d ", nstrats(pl));
	fprintf(sb, "\n");

	Profile	prof(this);
	save_game_format_iter(&prof, 0, sb, ucan);

	fclose(sb);
}


void	GameDescription::load_nfg_reduced(const char *fn, GameDescription *mn)
{
	zprintf("DDS vytvari vyslednou hru\n");
	copy_structure(*mn);
	parent = mn;

	int pl, sz, ke, ona;
	zprintf("DDS hra=%d hracu: \n", N);
	FILE *sb = fopen(fn, "rt");

	for (pl=0; pl<N; pl++) {
		fscanf(sb, "%d", &sz);
		zprintf("%d ", sz);
		for (ke=0; ke<sz; ke++) {
			fscanf(sb, "%d", &ona);
			inherited_strats[pl][ke] = ona;
			put_strategy(pl, mn->get_strat(pl, ona));
		}
	}
	zprintf("\n");

	fclose(sb);
}



void	GameDescription::map_back(longcis num, Profile *p_prof)
{
	p_prof->gamed = parent;
	Profile tmp(this);
	tmp.become_num(num);
	//zprintf("mapbek: "); tmp.printAddr(true); std::cout << "numje " << num << "\n";

	for (int a=0; a<N; a++) {
		//zprintf("%d ", inherited_strats[a][tmp.Addr[a]]);
		p_prof->Addr[a] = inherited_strats[a][tmp.Addr[a]];
	}

	p_prof->become_from_addr();
}


void	GameDescription::iter_profiles(Profiles &prfl, Ucache_manager &mng)
{
	Profile	addr(this);

	iter_profiles_iter(0, addr, prfl, mng);
}


void	GameDescription::iter_profiles_iter(int lev, Profile &addr, Profiles &prfl, Ucache_manager &mng)
{
	if (lev == N) {
		addr.become_from_addr();
		if (mng.exists(addr.num)) {
			Profile pom;
			if (mng.getp(addr.num))
				prfl.push_back(mng.getp(addr.num));
		}
		return ;
	}

	for (addr.Addr[lev]=0; addr.Addr[lev]<nstrats(lev); addr.Addr[lev]++) {
		iter_profiles_iter(lev+1, addr, prfl, mng);
	}

}

Ucache_manager_cm::~Ucache_manager_cm() 
{
	//if (cellM) delete cellM;	
}


Profile	*Ucache_manager_cm::get_seq(longcis num) 
{
	boost::mutex::scoped_lock lock(gsys_mut); 

	if (mapa.count(num)>0) {
		//zprintf("Znamy "); std::cout << num << "\n";
		return mapa[num];
	} else {
		// not known
		exit(1);
	
	}
}



Profile	*Ucache_manager_cm::get_seq_read(longcis num) 
{
	
	if (mapa.count(num)>0) {
		//zprintf("Znamy "); std::cout << num << "\n";
		return mapa[num];
	} else {
		return 0;
	
	}
	

	return mapa[num];
}


Profile	*Ucache_manager_cm::gsys_special_read(longcis num)
{
	boost::mutex::scoped_lock lock(gsys_mut); 

	if (mapa.count(num)>0) {
		//zprintf("Znamy "); std::cout << num << "\n";
		return mapa[num];
	} else {
		return 0;
	}
	

	return mapa[num];
}


int	Lexer::next()
{
	int zn = get();
	if (feof(sb)) return akt_lex=LEX_END;
	
	while (isspace(zn)) {
		zn = get();
		if (feof(sb)) return akt_lex=LEX_END;
	}

	char	*uk = id;
	bool	tec = false;

	//zprintf("ZN: %c\n", zn);

	if (isdigit(zn)) {
		num = 0;
		id[0]=0;
		while (isdigit(zn) || zn=='.') {
			*uk++=zn;
			zn = get();
			if (feof(sb)) break;
		}
		*uk=0;
		num = atof(id);
		put(zn);
		//zprintf("numero=%f\n", num);
		return akt_lex = LEX_NUM;
	}

	if (isalpha(zn)) {
		id[0]=0;
		while (isalpha(zn) || isdigit(zn)) {
			*uk++=zn;
			zn = get();
			//zprintf("ZNdal: %c\n", zn);
			if (feof(sb)) break;
		}
		*uk=0;
		put(zn);
		//zprintf("id=%s\n", id);
		return akt_lex = LEX_ID;
	}

	akt_lex = zn;
	return zn;
}



void	GVar_man::add(GVariable *gv)
{
	list.push_back(gv);
}


real	GVar_man::value(char *nm)
{
	GVariable *gv = find(nm);
	assert(gv!=0);
	return *gv->value;
}


void	GVar_man::read_config(char *fn)
{
	FILE *sb = fopen(fn, "rt");
	char st[100];
	float vl;

	while (!feof(sb)) {
		fscanf(sb, "%s %f", st, &vl);
		if (feof(sb)) break;
		GVariable *gva = find(st); assert(gva!=0);	
		*gva->value = vl;
	}

	fclose(sb);
}


GVariable *GVar_man::find(char *nm)
{
	for (GVary_i g=list.begin(); g!=list.end(); g++)
		if (!strcmp((*g)->nm, nm)) return *g;
	return 0;
}

