#ifndef	__GAMELIB_H
#define	__GAMELIB_H

typedef	void	cellModel;

//#pragma warning( disable : 4702 )

/*
#include <boost/array.hpp>
#include <boost/type.hpp>
#include <boost/multi_array.hpp>
*/

#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>

/*
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/io.hpp>
*/

#define	INIT_MERSENA

#include "mersennetwister.h"

#include <vector>
#include <set>
#include <map>
#include <list>
#include <string>



#include <gmpxx.h>
#include <assert.h>

#include <omp.h>

#include <stdio.h>

int	zprintf(const char *fmt, ...);

void	init_gamelib();

typedef	std::set<int>	Intset;
typedef	std::set<int>::iterator	Intset_i;

typedef	std::vector<int>	Intvector;
typedef	std::vector<int>::iterator	Intvector_i;

typedef	std::list<int>	Listvector;
typedef	std::list<int>::iterator	Listvector_i;


typedef	std::map<int,int>	IntInt_map;


//typedef	unsigned long long longcis;
typedef	mpz_class longcis;

struct ltstr_longcis
{
  bool operator()(const longcis s1, const longcis s2) const
  {
    return s1<s2;
  }
};

typedef	std::set<longcis, ltstr_longcis>	Longcisset;
typedef	std::set<longcis, ltstr_longcis>::iterator	Longcisset_i;


typedef	double	real;

typedef	std::map<int, real>	IntReal_map;
typedef	std::map<real, real>	RealReal_map;

typedef	std::set<real>	Realset;
typedef	std::set<real>::iterator	Realset_i;

typedef	std::vector<real>	Realvector;
typedef	std::vector<real>::iterator	Realvector_i;



//#include "kalendar.h"

class	GameDescription;

const	int	MAXplayers = 20;
const	int	MAXstrategyParams = 15;
const	int	MAXcpus = 20;

typedef	real	uvector[MAXplayers];


class	Strategy {
public:
	Strategy(GameDescription *g) {
		for (int a=0; a<MAXstrategyParams; a++) vals[a]=0;
		gamed=g;
		flag=false;
	}

	void	print();
	void	output(FILE *sb);
	bool	input(FILE *sb);

	real	vals[MAXstrategyParams];
	GameDescription	*gamed;
	int	size, order;

	bool	eq_except_one(int it, Strategy *s) {
		for (int a=0; a<size; a++) if (a!=it) 
			if (vals[a]!=s->vals[a]) return false;
		return true;
	}

	bool	eq(Strategy *s) {
		for (int a=0; a<size; a++) 
			if (vals[a]!=s->vals[a]) return false;
		return true;
	}

	bool	flag;
};

typedef	std::vector<Strategy *> Strategies;
typedef	std::vector<Strategy *>::iterator Strategies_i;


class	Profile {
public:
	Profile() {
	}

	Profile(GameDescription *g) { gamed=g; }

	void	operator = (Profile &prf) {
		copy(prf);
	}
	
	bool	operator == (Profile &prf);

	void	copy(Profile &prf);
	void	copy_u(Profile &prf);

	void	become_num(longcis eenum);
	void	become_from_addr();

	void	reset_u();

	void	load_strats();

	void	change_addr(int q, int n) {
		Addr[q]=n;
		become_from_addr();
	}

	int	Addr[MAXplayers];
	real	U[MAXplayers];
	Strategy *strats[MAXplayers];

	void	printAddr(bool nl=false, FILE *sb=stdout);
	bool	readAddr(FILE *sb);
	void	printU(bool nl=false, FILE *sb=stdout);
	void	printAll(bool nl=false, FILE *sb=stdout) {
		printAddr(false, sb);
		printU(nl, sb);
	}

	GameDescription	*gamed;
	longcis	num;
};



typedef	std::vector<Profile *> Profiles;
typedef	std::vector<Profile *>::iterator Profiles_i;


class	Entity {
public:
	Entity(int eNum) {
		eId = eNum;
	}

	int	eId;
};

typedef	std::vector<Entity *> Entities;
typedef	std::vector<Entity *>::iterator Entities_i;
typedef	std::map<int,Entity *> Entities_map;


class	Player {
public:
	Player(int pNum, Entity *ent=0) {	
		id = pNum;
		if (ent) addEntity(ent);
	}

	void	addEntity(Entity *en) { list.push_back(en); }

	//virtual	void	cm_message(cellModel *cm, int msg) {}

	int	id;
	Entities	list;
};



typedef	std::vector<Player *> Players;
typedef	std::vector<Player *>::iterator Players_i;
typedef	std::map<int,Player *> Players_map;


class	PlayerList {
public:
	PlayerList() {
	}

	void	add(Player *pl) {
		assert(mapa.count(pl->id)==0);
		id_to_num[pl->id] = list.size();
		list.push_back(pl);
		mapa[pl->id]=pl;
		ids.insert(pl->id);
	}

	Player	*get(int id) {
		return mapa[id];
	}

	int	pmap(int id) { 
		return id_to_num[id];
	}

	Players	list;
	Players_map mapa;
	Intset	ids;
	std::map<int,int> id_to_num;
};



enum	UCA_MSGS {
	UC_OK, UC_NONEX, UC_LOCKED, UC_OK_BAD
};

struct	Profile_num_i {
	Profile_num_i(longcis n, Profile *p) {
		num = n;
		prof = p;
	}

	longcis	num;
	Profile *prof;
};

typedef	std::vector<Profile_num_i> Profile_num_list;
typedef	std::vector<Profile_num_i>::iterator Profile_num_list_i;

struct	Ucache_manager {

	Ucache_manager() {
	}

	bool	exists(longcis num) {
		boost::mutex::scoped_lock lock(gsys_mut); 
		if (mapa.count(num)>0) {
			return true;
		} else {
			return false;
		}
	}

	int	get(longcis num, Profile *kam) {
		boost::mutex::scoped_lock lock(gsys_mut); 
		if (mapa.count(num)>0) {
			Profile *jeon = mapa[num];
			if (!jeon) return UC_OK_BAD;
			kam->copy_u(*jeon);
			return UC_OK;
		} else {
			if (the_lockeds.count(num)>0) return UC_LOCKED;
			the_lockeds.insert(num);
			return UC_NONEX;
		}
	}

	Profile	*getp(longcis num) {
		boost::mutex::scoped_lock lock(gsys_mut); 
		if (mapa.count(num)>0) {
			Profile *jeon = mapa[num];
			if (!jeon) return 0;
			return jeon;
		} else {
			return 0;
		}
	}

	void	put(longcis num, Profile *prf) {
		boost::mutex::scoped_lock lock(gsys_mut); 
		assert(the_lockeds.count(num)>0);
		the_lockeds.erase(num);

		if (prf) {
			Profile *np = new Profile(*prf);
			mapa[num] = np;
		} else mapa[num] = 0;
	}

	void	put_burst(Profile_num_list &ls) {
		boost::mutex::scoped_lock lock(gsys_mut); 
		for (Profile_num_list_i i=ls.begin(); i!=ls.end(); i++) {
			assert(the_lockeds.count(i->num)>0);
			the_lockeds.erase(i->num);

			mapa[i->num] = i->prof;
		}
	}

	void	put_seq(longcis num, Profile *prf) {
		boost::mutex::scoped_lock lock(gsys_mut); 

		//zprintf("Put_seq: "); std::cout<<num << " "; prf->printU(true);
		if (prf) {
			Profile *np = new Profile(*prf);
			mapa[num] = np;
			//mapa[num]->printU(true);
		} else mapa[num] = 0;
	}

	void	free_all() {
		for (std::map<longcis, Profile *>::iterator it=mapa.begin(); it!=mapa.end(); it++) {
			if (it->second) delete it->second;
		}
		mapa.clear();
	}

	boost::mutex    gsys_mut;
	std::map<longcis, Profile *> mapa;
	std::set<longcis>	the_lockeds;
};


struct	Ucache_manager_cm : public Ucache_manager {
	
	Ucache_manager_cm() { cellM=0; }

	Ucache_manager_cm(cellModel *cm) {
		cellM = cm;
	}

	~Ucache_manager_cm();

	Profile	*get_seq(longcis num);
	Profile	*get_seq_read(longcis num);

	Profile	*gsys_special_read(longcis num);

	void	*cellM;	
};

class	DDS_manager;

class	GameDescription {
public:
	GameDescription(int n=-1) {
		N=n;
		strategy_vector_size=0;
	}

	int	np() { return N; }
	int	nstrats(int q) { return strats_size[q]; }
	int	cnstrats(int q) { return strats[q].size(); }

	longcis	state_space_len() {
		longcis p = 1;
		for (int a=0; a<np(); a++) {
			p*=nstrats(a);
			//zprintf("je %lld\n", p);	
			//std::cout << p << "\n";
		}
		return p;
	}

	int	gnrows() {
		int n=0;
		for (int a=0; a<np(); a++) 
			n+= (nstrats(a) * (nstrats(a)-1));
		return n;
	}

	int	gncols() {
		int n=1;
		for (int a=0; a<np(); a++) 
			n *= nstrats(a);
		return n;
	}


	Strategy *get_strat(int q, int n) { return strats[q][n]; }

	void	put_strategies_n(int n, int hm) {
		for (int a=0; a<hm; a++) strats[n].push_back(new Strategy(this));
	}

	void	put_strategy(int n, Strategy *s) {
		strats[n].push_back(s);
	}

	void	prepare() { 
		map_init(); 
	}

	void	copy_structure(GameDescription &of) {
		N=of.N;
		for (Players_i p=of.players.list.begin(); p!=of.players.list.end(); p++)
			add_player(*p);
	}

	void	save_game_format_iter(Profile *prof, int level, FILE *sb, Ucache_manager &ucan);
	void	save_game_nfg(const char *fn, Ucache_manager &ucan);
	void	load_nfg_reduced(const char *fn, GameDescription *mn);
	void	load_nfg_reduced(DDS_manager *dds, GameDescription *mn);

	// addr->num maps
	longcis	idxs[MAXplayers];

	void	map_init();
	void	map2_prof(longcis num, Profile *kam);
	longcis	map2_num(Profile *kam);

	int	pmap(int id) { return players.pmap(id); }

	PlayerList	players;
	void	add_player(Player *pl) { players.add(pl); }
	Player	*get_player(int id) {
		return players.get(id);
	}

	IntInt_map	inherited_strats[MAXplayers];
	GameDescription	*parent;

	void	map_back(longcis num, Profile *p_prof);

	void	iter_profiles(Profiles &prfl, Ucache_manager &mng);
	void	iter_profiles_iter(int lev, Profile &addr, Profiles &prfl, Ucache_manager &mng);

	void	reset() {
		for (int a=0; a<MAXplayers; a++) 
			strats[a].clear();
	}

	int	N;
	Strategies strats[MAXplayers];
	int	strats_size[MAXplayers];
	int	strategy_vector_size;
};


extern	int playerCounter, entityCounter;


int	newPlayerId();
int	newEntityId();

real    myrand();
real    uniform(real f, real t);
int     int_uniform(real f, real t);

  
class   Stat {
public:
        Stat() { reset(); }

        void    reset() {
                min=100000000;
                max=-100000000;
                n=0;sum=0;
        }

        void    add(real l) {
            //    min5.push_back(l);
                if (l<min) { min=l; }
                if (l>max) { max=l;  }
                n++;
                sum+=l;
        }

	bool	valid() { return n>0; }
	real    avg() { return sum/(real)n; }
        void    operator()(real l) { add(l); }



        void    print() { zprintf("STAT: %.0f %.0f %.0f n=%d %.0f\n", min,max,avg(),n,sum); }

        real    min,max,sum;
        int     n;
};

struct	VazPrum {
	VazPrum() {
		cit=jm=0;
	}

	void	operator() (real c, real j) {
		cit += c*j;
		jm += j;
	}

	real	avg() { return cit/jm; }
	bool	valid() { return jm>0; }

	real	cit, jm;
};


real	procenta(real base, real pr);
real	vahovane(real m1, real c1, real m2, real c2);

enum	GP_CODES {
	GP_COMMENT,
	GP_MESSAGE
};

void	gprintf(int lev, const char *fmt, ...);

enum	{
	LEX_END = -1,
	LEX_ID	= 1001,
	LEX_NUM = 1002
};


class	Lexer {
public:
	Lexer(FILE *esb) {
		ret_zn = -1;
		sb = esb;
		next();
	}

	~Lexer() { fclose(sb); }

	int	get() { 
		if (ret_zn!=-1) { 
			int p = ret_zn; ret_zn=-1;
			return p;
		}

		int zn = fgetc(sb);
		if (zn=='#') {
			while (zn!='\n') {
				zn = fgetc(sb);
				if (feof(sb)) return EOF;
			}
		}

		return zn;
	}

	void	put(int zn) { 
		assert(ret_zn==-1);	
		ret_zn = zn;
	}

	int	next();
	void	expect(int zn) {	
		if (akt_lex!=zn) {
			zprintf("Chyba lexer %d != %d\n", akt_lex, zn);
			assert(akt_lex == zn);
		}
		
		next();
	}

	int	ret_zn, akt_lex;
	char	id[100];
	real	num;

	FILE	*sb;
};


struct	GVariable {
	GVariable(const char *name, real *v) {
		nm = strdup(name);
		value = v;
	}

	GVariable(const char *name, int *v) {
		nm = strdup(name);
		valuei = v;
	}

	real	*value;
	int	*valuei;
	char	*nm;
};

typedef	std::vector<GVariable *> GVary;
typedef	std::vector<GVariable *>::iterator GVary_i;


class	GVar_man {
public:
	GVar_man() {
	}

	void	add(GVariable *gv);
	real	value(char *nm);
	void	read_config(char *fn);
	GVariable *find(char *nm);

	GVary	list;
};


struct	Interval {
	Interval(real ff, real ft, real fs, real fmins=0) {
		from=ff; to=ft; step = std::max(fs, fmins);
		real val;
		for (val=from; val<to; val+=step)
			list.push_back(val);
		if (val>to) list.push_back(to);
	}


	real	from, to, step, act;
	bool	clos;
	Realvector	list;
};

#endif

