#include "gsys-aumann.h"


void	CorrSolver::create_vary(Profile *prof, int lev, LP_vary &vary)
{
	if (lev==gamed->N) {
		prof->become_from_addr();
		Profile *res = ucache->get_seq_read(prof->num);
		//assert(ucache->exists(prof->num));
		
		if (res) {
			LPvar_cs *nj = new LPvar_cs(res);
			vary.push_back(nj);
			if (uZ(nj)==0) {	
				nu++;
			} else nenu++;
		}

		return ;
	}

	for (prof->Addr[lev]=0; prof->Addr[lev]<gamed->nstrats(lev); prof->Addr[lev]++) {
		create_vary(prof, lev+1, vary);
	}
}



bool	CorrSolver::work()
{
	LP_solver lps(LPX_MAX, -1);
	LP_vary vary, varyr;
	Profile prof(gamed);

	if (texto) zprintf("CorrSolver: creating profiles\n");
	create_vary(&prof, 0, vary);

	if (texto) {
		zprintf("======================================================================\n");
		zprintf("%d LP-variables empty/non-empty %d/%d profiles\n", vary.size(), nu, nenu);
		zprintf("======================================================================\n");
	}

	LP_vary_i li;
	int	pl;
	int	idxa;
	bool	optimaz = true;
	
	LP_nerovny nerovny;
	LPvar_cs *lpcs;

	#pragma	omp parallel for schedule(dynamic, 1000) private(idxa) num_threads(MAXcpus)
	for (idxa=0; idxa<vary.size(); idxa++) {
		lpcs = (LPvar_cs*) vary[idxa];
		lpcs->z_coef = uZ(lpcs)/divider;
	}

	/*	
	#pragma omp parallel for schedule(dynamic,1) private(pl) num_threads(MAXcpus)
	for (pl=0; pl<gamed->N; pl++) {
	}
	*/

	int celkem_itemu=0;
	int	mlp_rad=0;

	#pragma omp parallel for schedule(dynamic,1) private(pl) num_threads(MAXcpus)
	for (pl=0; pl<gamed->N; pl++) {
		LPvar_cs *lpp;
		real	zs[2], roz;
		int	ods, dos;
		int	klad,zap;
		int	idx;
	
		if (texto) zprintf("Generating player %d, %d strategies\n", pl, gamed->nstrats(pl));
		for (ods=0; ods<gamed->nstrats(pl); ods++) 
		//for (dos=0; dos<gamed->nstrats(pl); dos++) if (ods!=dos) {
		for (dos=gamed->nstrats(pl)-1; dos>=0; dos--) if (ods!=dos) {
			LP_nerovn *nro = new LP_nerovn(LPX_LO, 0, 0);
			if (texto) zprintf("G-matrix %d %d %d: ", pl, ods, dos); klad=zap=0;

			for (idx=0; idx<vary.size(); idx++) {
			lpp = (LPvar_cs*) vary[idx];
			if (lpp->probable) {
				if (lpp->prof->Addr[pl]==ods) {
					Profile nprof(*lpp->prof);
					zs[0] = lpp->prof->U[pl];
					nprof.Addr[pl]=dos;
					nprof.become_from_addr();
					Profile *res = ucache->get_seq(nprof.num);
					if (res) {
						zs[1] = res->U[pl];
						if (zs[0]!=zs[1]) {
							roz = (zs[0] - zs[1])/divider;
							nro->of.push_back(LP_ref(lpp,roz));
							if (roz>0) klad++;
							if (roz<=0) zap++;
						}
					}
				}
			}}

			if (texto) zprintf("| %d %d [%d]\n", klad, zap, celkem_itemu);

			#pragma omp critical
			{
				if (nro->of.size()>0 && klad>0) {
					nerovny.push_back(nro);
					celkem_itemu += (klad+zap);
					mlp_rad++;
				}

				if (nro->of.size()>0 && zap>0 && klad==0) {
					for (LP_refy_i ll=nro->of.begin(); ll!=nro->of.end(); ll++) {
						if (optimaz) ((LPvar_cs *)ll->var)->probable=false;
						ll->var->ido=0;
					}
				}
			}
		}
	}
	
	int	idx;

	Intset	zbylojim[MAXplayers];
	for (li=vary.begin(); li!=vary.end(); li++) {
		lpcs = (LPvar_cs*) *li;
		if (lpcs->probable) {
			varyr.push_back(*li);
			for (pl=0; pl<gamed->N; pl++) 
				zbylojim[pl].insert(lpcs->prof->Addr[pl]);
		} else lpcs->z_coef=0;
	}

	zprintf("Reduced game: \n");
	for (pl=0; pl<gamed->N; pl++) {
		zprintf("Player-%d has the resulting strategies: ", pl);
		for (Intset_i i=zbylojim[pl].begin(); i!=zbylojim[pl].end(); i++)
			zprintf("%d ", *i);
		zprintf("\n");
	}

	for (li=varyr.begin(); li!=varyr.end(); li++) {
		lpcs = (LPvar_cs*) *li;
		//zprintf("%d ", lpcs->id);
	}

	LPvar_cs *lp_best = 0;
	real	maxz = -10000000000.0;
	int	kolmx = 0;
	
	bool	solve_lp = true;

	if (mlp_rad<=1) {
		zprintf("CESolver with a single-row G-matrix => it finds the most efficient profile.\n");

		for (idx=0; idx<varyr.size(); idx++) {
			lpcs = (LPvar_cs*) varyr[idx];
			if (lpcs->z_coef>maxz) { lp_best=lpcs; maxz=lpcs->z_coef; kolmx=1; }
			else
			if (lpcs->z_coef==maxz) { kolmx++; }
		}

		zprintf("The most efficient profit %.0f; there are %d such profiles in the game.\n", maxz, kolmx);

		if (kolmx>1) {
			for (idx=0; idx<varyr.size(); idx++) {
				lpcs = (LPvar_cs*) varyr[idx];
				if (lpcs->z_coef==maxz) { 
					lpcs->value = 1.0/(real)kolmx;
				}
			}
		} else
			lp_best->value = 1;

		solve_lp = false;
		if (celkem_itemu==0) celkem_itemu = varyr.size();
	}


	LP_nerovn *jedna = new LP_nerovn(LPX_FX, 1, 1);
	for (li=varyr.begin(); li!=varyr.end(); li++) {
		jedna->of.push_back(LP_ref(*li, 1));
	}

	for (li=varyr.begin(); li!=varyr.end(); li++) 
		lps.put_var(**li);

	celkem_itemu=0;
	int nz, jz;

	int nr=0;
	for (LP_nerovny_i ne=nerovny.begin(); ne!=nerovny.end(); ne++) {
		nz = (*ne)->nonzeros();
		jz = (*ne)->nzeros();
		if (nz>0) {
			lps.put_nerovn(*ne);
			celkem_itemu += nz;
			nr++;
			//zprintf("Radek %d bude mit %d/%d/%d\n", nr, nz, jz, (*ne)->nn());
		} 
	}

	lps.put_nerovn(jedna);
	lps.init_mspace(celkem_itemu + varyr.size() + 1);


	if (texto) zprintf("lps.work; %d constraints, %d lp-variables %d non-zero matrix elements\n", lps.nerovny.size(), varyr.size(), celkem_itemu);

	lps.flag_out=true;
	lps.init_rows_cols();

	if (solve_lp) {
		lps.work();
	}

PRIMY_KONEC:


	if (texto) zprintf("CE-Probability distribution\n");
	for (li=varyr.begin(); li!=varyr.end(); li++) {
		if ((*li)->value>0.01) {
			if (texto) {
				zprintf("Lp-Variable(profile number): %d %.2f\n", (*li)->id, (*li)->value);
				((LPvar_cs *)(*li))->prof->printAll(true);
			}
			rnd.insert(((LPvar_cs*)(*li))->prof->num, (*li)->value);
		}
	}

	solution_p = rnd.maxprof();

        if (texto) { zprintf("Solution-profile: "); std::cout << solution_p << "\n"; }
        Profile last(gamed);
        last.become_num(solution_p);

	return true;
}


void	CorrSolver::export_solution(CorrSolution &sol)
{
	sol.solution_p = solution_p;
	for (PR_itemy_i r=rnd.itemy.begin(); r!=rnd.itemy.end(); r++) {
		sol.itemy.push_back(*r);
	}
}

