#ifndef	__OPTIM_H
#define	__OPTIM_H

#include <stdio.h>

#include "gamelib.h"
#include "mujpool.h"
#include "mersennetwister.h"

#define	MUSI_SKONCIT_DOBRE

class	LP_solver;

extern	"C" {
#include <glpk.h>
			 
#undef	print
#undef	dzprintf
}


#include <vector>

class	LP_var {
public:
	LP_var() { id = -123456;}
	
	void	gener(LP_solver &slv);
	void	svalue(float v);

	void	sb_export(FILE *sb) {
		fprintf(sb, "%.0f ", z_coef);
	}

	void	set(real o, real d, real z) {
		iod = o; ido=d; z_coef=z;
	}
	
	int	id;
	float	z_coef, value, iod, ido;
};


typedef	std::vector<LP_var *> LP_vary;
typedef	std::vector<LP_var *>::iterator LP_vary_i;


class	LP_ref {
public:
	LP_ref(LP_var *evar, float co) {
		var=evar;
		coef=co;
	};
	
	LP_var	*var;
	float	coef;
};


#define	LPX_MINIMALIZUJ	true
#define	LPX_MAXIMALIZUJ false


typedef	std::vector<LP_ref> LP_refy;
typedef	std::vector<LP_ref>::iterator LP_refy_i;


class	LP_nerovn {
public:
	LP_nerovn() {
		iod=ido=0;
		cd = LPX_DB;
	}

	LP_nerovn(int ecd, float eiod, float eido) {
		cd=ecd; iod=eiod; ido=eido;
	};

	void	set_constr(int ecd, float eiod, float eido) {
		cd=ecd; iod=eiod; ido=eido;
	}

	void	reset() { of.clear(); }
	
	void	gener(LP_solver &slv);

	void	sb_export(int cols, FILE *sb) {
		int kde=1, i;
		for (LP_refy_i l=of.begin(); l!=of.end(); l++) {
			while (l->var->id<kde) {
				fprintf(sb, "0 ");
				kde++;
			}
			fprintf(sb, "%.0f ", l->coef); kde++;
		}

		while (kde<=cols) {
			fprintf(sb, "0 ");
			kde++;
		}

		fprintf(sb, "\n", ido);
	}

	int	nonzeros() {
		int s=0;
		for (LP_refy_i l=of.begin(); l!=of.end(); l++) 
			if (l->var->z_coef!=0) s++;
		return s;
	}

	int	nzeros() {
		int s=0;
		for (LP_refy_i l=of.begin(); l!=of.end(); l++) 
			if (l->var->z_coef==0) s++;
		return s;
	}

	int	nn() { return of.size(); }
	
	float	ido, iod;
	int	radek;
	int	cd;
	LP_refy	of;
};

typedef	Mujpul<LP_nerovn> LPnerovn_pool;

typedef	std::vector<LP_nerovn *> LP_nerovny;
typedef	std::vector<LP_nerovn *>::iterator LP_nerovny_i;

extern	int	LP_solv_cc;

const int matr_size=60001;
//const int matr_size=2000000;

class	LP_solver {
public:
	LP_solver(int tp=LPX_MIN, int enumer=matr_size) {
		dir=tp;
		flag_out=false;
		Mipjo=false;
		open_prob();
		LP_solv_cc++;
		texto=false;

		if (enumer!=-1) init_mspace(enumer);
	}

	void	init_mspace(int enumer) {
		numero = enumer;
		ia = new int[numero];
		ja = new int[numero];
		ar = new double[numero];
	}

	~LP_solver() {
		delete [] ia;
		delete [] ja;
		delete [] ar;

		close_prob();
	}
	
	void	open_prob() {
		lp = lpx_create_prob();
		//lpx_set_class(lp, LPX_MIP);
		//lpx_set_prob_name(lp, "sample electric");
	
		lpx_set_obj_dir(lp, dir);	
		//lpx_set_int_parm(lp, LPX_K_MSGLEV, 0);
		vars=ida=0;
		rads=0;
		/*
		for (int a=0; a<1001; a++) {
			ia[a]=ja[a]=0;
			ar[a]=0;
		}
		*/
		for (int a=0; a<10; a++)
			ners[a]=0;
	
	}
	
	void	close_prob() {
		lpx_delete_prob(lp);
		vary.clear();
		nerovny.clear();
	}
	
	
	
	int	new_var() { return ++vars; }
	int	new_ida() { 
		int a=++ida; 
		//zprintf("IDa %d/%d \n", ida, numero);
		assert(a<numero);
		return a;
	}
	int	new_rad() { return ++rads; }

	void	put_var(LP_var &vr) {
		vr.id = new_var();
		vary.push_back(&vr);
	}

	void	put_nerovn(LP_nerovn *nr) {
		nr->radek = new_rad();
		nerovny.push_back(nr);
	}

	void	sb_export(FILE *sb) {
		fprintf(sb, "%d %d\n", rads-1, vary.size());
		for (LP_nerovny_i n=nerovny.begin(); n!=nerovny.end(); n++)
			(*n)->sb_export(vary.size(), sb);
		for (LP_vary_i v=vary.begin(); v!=vary.end(); v++)
			(*v)->sb_export(sb);
		fprintf(sb, "\n");
	}
	
	void	init_rows_cols() {
		//zprintf("Rows,Cols: %d %d nerovny.size=%d\n", rads, vary.size(), nerovny.size());
		lpx_add_rows(lp, rads);
		lpx_add_cols(lp, vary.size());
		for (LP_vary_i v=vary.begin(); v!=vary.end(); v++)
			(*v)->gener(*this);
		for (LP_nerovny_i n=nerovny.begin(); n!=nerovny.end(); n++)
			(*n)->gener(*this);
	}
	
	void	matrix_put(int id, int i, int j, float ra) {
		ia[id]=i;
		ja[id]=j;
		ar[id]=ra;
	}
	
	void	matrix_put_new(int i, int j, float ra) {
		int id=new_ida();
		//zprintf("Matrix[%d] %d %d %f\n", id, i, j, ra);
		ia[id]=i;
		ja[id]=j;
		ar[id]=ra;
		//int k = id;
		//zprintf("->(%d) %d %d %.0f\n", ida, ia[k], ja[k], ar[k]);
	}
	
	bool	work();
	
	void	print_res(FILE *sb=stdout) {
		fprintf(sb, "RES ");
		//for (int a=0; a<vars; a++) 
		//	fprintf(sb, "%.0f ", results[a]);
		fprintf(sb, "\n");
	}
	
	LPX *lp;
	int	dir;
	bool flag_out, Mipjo;

	//int ia[matr_size], ja[matr_size];
	//double ar[matr_size];
	int	*ia, *ja;
	double	*ar;

	//float	results[1+100000];
	
	int	vars, ida, rads;
	LP_vary	vary;
	LP_nerovny nerovny;
	
	LP_nerovn *ners[10];

	int	numero;
	
	float	Z;

	bool	texto;
};


struct	PR_item {
	PR_item(longcis i, float f) : key(i), prob(f) {}
	longcis	key;
	float	prob;
} ;
	

typedef	std::vector<PR_item>	PR_itemy;
typedef	std::vector<PR_item>::iterator	PR_itemy_i;

class	ProgRND {
public:
	ProgRND() {
		p_sum=0;
	}

	void	insert(longcis key, float prob) {
		itemy.push_back(PR_item(key, prob));
		p_sum+=prob;
	}

	float	mrand() { return genrand_real1(); }

	longcis	gener() {
		float kam = mrand();
		float cum = 0;
		//dzprintf("Gener %.4f\n", kam);
		for (std::vector<PR_item>::iterator i=itemy.begin(); i!=itemy.end(); i++) {
			if (kam>=cum && kam<=cum+i->prob) return i->key;
			cum+=i->prob;
		}
		return -1;
	}

	longcis	maxprof() {
		real prba = 0;
		longcis kde=0;
		for (std::vector<PR_item>::iterator i=itemy.begin(); i!=itemy.end(); i++) {
			if (i->prob > prba) {
				kde = i->key;
				prba = i->prob;
			}
		}
		return kde;
	}

	PR_itemy itemy;
	float	p_sum;
};



#endif

